local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
Sackville.Section = {}
Sackville.Section.__index = Sackville.Section

local function MakeTitleFrame(section)
    local basename = string.format("Sackville_%s_%s", 
        section.bagName, section.name)

    local titleFrame = _G[basename .. "_TitleFrame"] or 
        CreateFrame("Frame", basename .. "_TitleFrame", 
            Sackville:GetBag(section.bagName):GetFrame())

    titleFrame.texture = titleFrame:CreateTexture()
    titleFrame.texture:SetAllPoints(titleFrame)
    titleFrame.texture:SetTexture(0, 0.3, 0, 1)
            
    local text = _G[basename .. "_TitleText"] or 
        titleFrame:CreateFontString(basename .. "_TitleText")
    text:SetFontObject("GameFontWhite")
    text:SetText(section.name)
    text:ClearAllPoints()
    text:SetPoint("TOPLEFT")
    text:SetPoint("TOPRIGHT")
    text:SetHeight(Sackville.Options.section.title_height)
    text:SetJustifyH("LEFT")

    titleFrame.text = text
    titleFrame.section = section
    -- titleFrame:SetWidth(text:GetStringWidth())
    titleFrame:SetHeight(Sackville.Options.section.title_height)
    titleFrame:Show()
    titleFrame:SetScript("OnMouseUp", function(self)
        self.section:TitleClicked() end)

    return titleFrame
end

function Sackville.Section:New(parentBag, sectionName, sectionConfig)
    Sackville:Debug(1, "section", "Adding new section: %s", sectionName)
    local self = {}
    setmetatable(self, Sackville.Section)
    
    local section = {}
    local sectionDefaults = Sackville.Options.section
    self.name = sectionName
    self.bagName = parentBag.name
    self.displayOrder = sectionConfig.displayOrder
    if sectionConfig.useDefaultPriority then
        self.priority = parentBag.defaultPriority
    else
        self.priority = sectionConfig.priority
    end
    if sectionConfig.useDefaultSorting then
        self.sortCriteria = parentBag.defaultSortCriteria
    else
        self.sortCriteria = sectionConfig.sortCriteria
    end
    
    self.alwaysMatch = sectionConfig.alwaysMatch or false
    self.categories = sectionConfig.categories or {}
    self.showEmptySlot = sectionConfig.showEmptySlot or false
    self.alwaysShow = sectionConfig.alwaysShow
    self.minimized = false
    self.itemKeys = {}
    self.titleFrame = MakeTitleFrame(self)
    return self
end

function Sackville.Section:IsShown()
    return self.alwaysShow or self:HasItems()
end

function Sackville.Section:ShowTitle()
    self.titleFrame:Show()
end

function Sackville.Section:HideTitle()
    self.titleFrame:Hide()
end

function Sackville.Section:SetTitlePoint(...)
    self.titleFrame:ClearAllPoints()
    self.titleFrame:SetPoint(...)
end

function Sackville.Section:TitleClicked()
    if not self.alwaysShow then
        self.minimized = not self.minimized
        bag = Sackville:GetBag(self.bagName)
        bag:LayoutItems()
        Sackville:LayoutBags(bag)
    end
end

function Sackville.Section:UpdateTitle(text)
    self.titleFrame.text:SetText(text)
    if self.minimized then
        self.titleFrame.texture:Show()
    else
        self.titleFrame.texture:Hide()
    end
end

function Sackville.Section:UpdateTitleWidth(newWidth)
    -- Sackville:Debug(1, "section", "UpdateTitleWidth(%d) => %d", maxWidth, newWidth)
    self.titleFrame:SetWidth(newWidth)
    -- self.titleFrame.text:SetWidth(newWidth)
end

function Sackville.Section:GetTitleStringWidth()
    return self.titleFrame.text:GetStringWidth()
end

function Sackville.Section:AddItem(item)
    item:SetBagName(self.bagName)
    item:SetSectionName(self.name)
    item:SetParent(self.frame)
    item:Show()
    
    table.insert(self.itemKeys, item:GetBagSlot())

    Sackville:Debug(2, "section", "AddItem: %s", item:Dump())
end

function Sackville.Section:RemoveItem(item)   
    Sackville:Debug(2, "section", "RemoveItem: %s", item:Dump())
    item:SetBagName("")
    item:SetSectionName("")
    item:Hide()
    item:SetParent(Sackville.dummyFrame)
    
    for i=1,#self.itemKeys do
        if self.itemKeys[i] == item:GetBagSlot() then
            table.remove(self.itemKeys, i)
            break
        end
    end
end

function Sackville.Section:HasItems()
    return Sackville.Util.TableHasValues(self.itemKeys)
end

function Sackville.Section:ItemIterator()
    local myItems = {}
    for i, bagslot in ipairs(self.itemKeys) do
        myItems[bagslot] = Sackville.db.char.items[bagslot]
    end
    return Sackville.Util.kpairs(myItems, self.sortCriteria)
end

function Sackville.Section:MatchItem(item)
-- Returns true if the item matches the categories assigned to this section
    if self.alwaysMatch then
        Sackville:Debug(3, "filter", "* Hit default section: %s", self.name)
        return true
    end
    
    
    for _, catName in ipairs(self.categories) do
        local matchRules = Sackville.Options.categories[catName]

        if matchRules and #matchRules > 0 then
            local isMatch = Sackville.Filter:ItemMatchesRules(item, matchRules)
            if isMatch then return true end
        else
            Sackville:Debug(4, "filter", "* Found no rules in section: %s", self.name)        
        end
    end
    
    return false    
end

function Sackville.Section:HideAll(bagType)
    for _, item in self:ItemIterator() do
        item:Hide()
    end
    self:HideTitle()
    local emptyItem = self:UpdateEmptySlot(bagType)
    if emptyItem then emptyItem:Hide() end
end

function Sackville.Section:LayoutContents(bag)
    -- TODO: Support minimization
    Sackville:Debug(1, "section", "LayoutContents for section %s", self.name)
 
    local start_x, start_y, end_x 
    
    if self.alwaysShow or self:HasItems() then
        if self.minimized then
            self:UpdateTitle(string.format("+ %s (%d)", self.name, #self.itemKeys))
            self:UpdateTitleWidth(self:GetTitleStringWidth())
            self:HideAll()
        else
            self:UpdateTitle(self.name)
            start_x, start_y, end_x = self:PositionItems(bag)  
            local x, y = self:UpdateEmptySlot(bag)
            start_x = start_x or x
            start_y = start_y or y
            end_x = end_x or x
            self:UpdateTitleWidth(math.min(self:GetTitleStringWidth(),
                end_x - start_x + Sackville.Options.BUTTON_SIZE + 
                (Sackville.Options.section.padding_right / 2)))
            self:ShowTitle()
            self:SetTitlePoint("TOPLEFT", start_x, 
                start_y + Sackville.Options.section.margin_top)
        end
    else
        self:HideTitle()
    end
    
    return self.minimized
end

function Sackville.Section:PositionItems(bag)
    local frame = bag:GetFrame()  
    local start_x, start_y, end_x
    for _, item in self:ItemIterator() do
        local x, y = bag:GetNextItemPosition()
        Sackville:Debug(3, "section", "Got next item pos: (%d,%d)", x, y)
        if start_x == nil then start_x = x end
        if start_y == nil then start_y = y end
        if start_y == y then end_x = x end
        
        item:SetParent(frame)
        item:SetPoint("TOPLEFT", x, y)
        item:Show()
    end
    
    return start_x, start_y, end_x
end

function Sackville.Section:HighlightSearchItems(text)
    text = string.lower(text)
    for _, item in self:ItemIterator() do
        if text == "" then
            item.button:SetAlpha(1)
        elseif string.find(string.lower(item.name), text) then
            item.button:SetAlpha(1)
        else
            item.button:SetAlpha(0.25)
        end
    end
end

function Sackville.Section:UpdateEmptySlot(bag)            
    if not self.showEmptySlot then return nil; end
    local emptyCount = Sackville:GetEmptySlotCount(bag.bagType)
    if emptyCount == 0 then return end
    
    local bagSet = nil   
      
    local emptyItem = Sackville.db.char.emptyItem[bag.bagType]
    
    for bagslot, item in pairs(Sackville.db.char.items) do
        if item:IsEmpty() and Sackville:IsBagType(bag.bagType, item:GetBagId()) then
            if emptyItem == nil or not emptyItem.empty then
                emptyItem = item
            else
                item:Hide()
            end
        end
    end

    if emptyItem then
        emptyItem:SetButtonCount(Sackville:GetEmptySlotCount(bag.bagType))
        Sackville.db.char.emptyItem[bag.bagType] = emptyItem
    end
    
    if self.showEmptySlot and emptyItem then
        local frame = bag:GetFrame()
        emptyItem:Show()
        emptyItem:SetParent(frame)
        local x, y = bag:GetNextItemPosition()    
        emptyItem:SetPoint("TOPLEFT", x, y)
        return x, y
    end
end

function Sackville.Section:Show()
    self.frame:Show()
end

function Sackville.Section:Hide()
    self.frame:Hide()
end

function Sackville.Section:GetName()
    return self.name
end

function Sackville.Section:SetMaxWidth(width)
    self.max_width = width
end
